"""Lidarr Models."""
# pylint: disable=invalid-name, too-many-instance-attributes
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum

from ..const import ARTIST_ID, DATE, PATH, TITLE
from .base import BaseModel
from .lidarr_common import (
    _LidarrAddOptions,
    _LidarrAlbumCommon,
    _LidarrArtist,
    _LidarrArtistTitleInfo,
    _LidarrAudioTags,
    _LidarrCommon,
    _LidarrCommon3,
    _LidarrCommon5,
    _LidarrFields,
    _LidarrImportListPreset,
    _LidarrMediaInfo_Quality,
)
from .request_common import (
    _Common3,
    _Common4,
    _Common7,
    _Common8,
    _Editor,
    _HistoryData,
    _IsLoaded,
    _ManualImport,
    _Monitor,
    _MonitorOption,
    _Quality,
    _Ratings,
    _RecordCommon,
    _ReleaseCommon,
    _Rename,
    _RetagChange,
    _StatusMessage,
    _TagDetails,
)


class LidarrCommands(str, Enum):
    """Lidarr commands."""

    ALBUM_SEARCH = "AlbumSearch"
    APP_UPDATE_CHECK = "ApplicationUpdateCheck"
    ARTIST_SEARCH = "ArtistSearch"
    DOWNLOADED_ALBUMS_SCAN = "DownloadedAlbumsScan"
    MISSING_ALBUM_SEARCH = "MissingAlbumSearch"
    REFRESH_ALBUM = "RefreshAlbum"
    REFRESH_ARTIST = "RefreshArtist"


class LidarrEventType(Enum):
    """Lidarr event types."""

    DELETED = 5
    DOWNLOAD_FAILED = 4
    DOWNLOAD_IMPORTED = 8
    GRABBED = 1
    IGNORED = 7
    IMPORT_FAILED = 7
    RENAMED = 6
    RETAGGED = 9
    TRACK_IMPORTED = 3


class LidarrImportListActionType(str, Enum):
    """Lidarr import list action types."""

    GET_PLAYLISTS = "getPlaylists"
    GET_PROFILES = "getProfiles"
    GET_TAGS = "getTags"


class LidarrImportListType(str, Enum):
    """Lidarr import list type."""

    LAST_FM = "lastFm"
    OTHER = "other"
    PROGRAM = "program"
    SPOTIFY = "spotify"


class LidarrImportListMonitorType(str, Enum):
    """Lidarr import list monitor type."""

    ENTIRE_ARTIST = "entireArtist"
    NONE = "none"
    SPECIFIC_ALBUM = "specificAlbum"


class LidarrSortKeys(str, Enum):
    """Lidarr sort keys."""

    ALBUM_TITLE = "albums.title"
    ARTIST_ID = ARTIST_ID
    DATE = DATE
    DOWNLOAD_CLIENT = "downloadClient"
    ID = "id"
    INDEXER = "indexer"
    MESSAGE = "message"
    PATH = PATH
    PROGRESS = "progress"
    PROTOCOL = "protocol"
    QUALITY = "quality"
    RATINGS = "ratings"
    RELEASE_DATE = "albums.releaseDate"
    SOURCE_TITLE = "sourcetitle"
    STATUS = "status"
    TIMELEFT = "timeleft"
    TITLE = TITLE


@dataclass(init=False)
class LidarrAlbum(_LidarrCommon3, _LidarrAlbumCommon):
    """Lidarr album attributes."""

    added: datetime
    addOptions: type[_LidarrAddOptions] = field(default=_LidarrAddOptions)
    albumReleases: type[_IsLoaded] = field(default=_IsLoaded)
    artistMetadata: type[_IsLoaded] = field(default=_IsLoaded)
    artistMetadataId: int
    cleanTitle: str
    lastInfoSync: datetime
    oldForeignAlbumIds: list[str]

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.addOptions = _LidarrAddOptions(self.addOptions)
        self.artistMetadata = _IsLoaded(self.artistMetadata)
        self.albumReleases = _IsLoaded(self.albumReleases)


@dataclass(init=False)
class LidarrArtist(_LidarrArtist):
    """Lidarr artist attributes."""

    lastAlbum: type[LidarrAlbum] = field(default=LidarrAlbum)
    nextAlbum: type[LidarrAlbum] = field(default=LidarrAlbum)

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.lastAlbum = LidarrAlbum(self.lastAlbum)
        self.nextAlbum = LidarrAlbum(self.nextAlbum)


@dataclass(init=False)
class LidarrAlbumEditor(BaseModel):
    """Lidarr album editor attributes."""

    albumids: list[int]
    monitored: bool


@dataclass(init=False)
class LidarrAlbumLookup(_LidarrAlbumCommon):
    """Lidarr album lookup attributes."""

    remoteCover: str


@dataclass(init=False)
class LidarrBlocklistItem(_Common7, _LidarrCommon5):
    """Lidarr blocklist item attributes."""

    albumIds: list[int]
    message: str


@dataclass(init=False)
class LidarrBlocklist(_RecordCommon):
    """Lidarr blocklist attributes."""

    records: list[LidarrBlocklistItem] = field(
        default_factory=list[LidarrBlocklistItem]
    )

    def __post_init__(self):
        """Post init."""
        self.records = [LidarrBlocklistItem(record) for record in self.records]


@dataclass(init=False)
class LidarrCalendar(LidarrAlbum):
    """Lidarr calendar attributes."""


@dataclass(init=False)
class LidarrWantedCutoff(_RecordCommon):
    """Lidarr wanted cutoff attributes."""

    records: list[LidarrAlbum] = field(default_factory=list[LidarrAlbum])

    def __post_init__(self):
        """Post init."""
        self.records = [LidarrAlbum(record) for record in self.records]


@dataclass(init=False)
class LidarrImportList(_Common3, _LidarrImportListPreset):
    """Lidarr import list attributes."""

    fields: list[_LidarrFields] | None = None
    implementationName: str
    monitorNewItems: str
    presets: list[_LidarrImportListPreset] | None = None
    rootFolderPath: str
    shouldMonitorExisting: bool
    shouldSearch: bool

    def __post_init__(self):
        """Post init."""
        self.fields = [_LidarrFields(field) for field in self.fields or []]
        self.presets = [_LidarrImportListPreset(x) for x in self.presets or []]


@dataclass(init=False)
class LidarrAlbumHistory(_LidarrCommon, _LidarrCommon5):
    """Lidarr album history attributes."""

    album: type[LidarrAlbum] = field(default=LidarrAlbum)
    data: type[_HistoryData] = field(default=_HistoryData)
    downloadId: str
    eventType: str
    trackId: int

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.album = LidarrAlbum(self.album)
        self.data = _HistoryData(self.data)


@dataclass(init=False)
class LidarrHistory(_RecordCommon):
    """Lidarr history attributes."""

    records: list[LidarrAlbumHistory] = field(default_factory=list[LidarrAlbumHistory])

    def __post_init__(self):
        """Post init."""
        self.records = [LidarrAlbumHistory(field) for field in self.records]


@dataclass(init=False)
class _LidarrAlbumType(BaseModel):
    """Lidarr album type attributes."""

    albumType: type[_Common3] = field(default=_Common3)
    allowed: bool

    def __post_init__(self):
        """Post init."""
        self.albumType = _Common3(self.albumType)


@dataclass(init=False)
class _LidarrReleaseStatus(BaseModel):
    """Lidarr release status attributes."""

    allowed: bool
    releaseStatus: type[_Common3] = field(default=_Common3)

    def __post_init__(self):
        """Post init."""
        self.releaseStatus = _Common3(self.releaseStatus)


@dataclass(init=False)
class LidarrMetadataProfile(_Common3):
    """Lidarr metadata profile attributes."""

    primaryAlbumTypes: list[_LidarrAlbumType] | None = None
    releaseStatuses: list[_LidarrReleaseStatus] | None = None
    secondaryAlbumTypes: list[_LidarrAlbumType] | None = None

    def __post_init__(self):
        """Post init."""
        self.primaryAlbumTypes = [
            _LidarrAlbumType(x) for x in self.primaryAlbumTypes or []
        ]
        self.releaseStatuses = [
            _LidarrReleaseStatus(x) for x in self.releaseStatuses or []
        ]
        self.secondaryAlbumTypes = [
            _LidarrAlbumType(x) for x in self.secondaryAlbumTypes or []
        ]


@dataclass(init=False)
class _LidarrQueueItemAlbum(_LidarrAlbumCommon):
    """Lidarr queue detail album attributes."""

    id: int


@dataclass(init=False)
class LidarrQueueItem(_Common4, _Common7, _Common8):
    """Lidarr queue item attributes."""

    album: type[_LidarrQueueItemAlbum] = field(default=_LidarrQueueItemAlbum)
    albumId: int
    artist: type[_LidarrArtist] = field(default=_LidarrArtist)
    artistId: int
    downloadForced: bool

    def __post_init__(self):
        """Post init."""
        self.album = _LidarrQueueItemAlbum(self.album)
        self.artist = _LidarrArtist(self.artist)
        self.quality = _Quality(self.quality)
        self.statusMessages = [_StatusMessage(x) for x in self.statusMessages or []]


@dataclass(init=False)
class LidarrQueue(_RecordCommon):
    """Lidarr queue attributes."""

    records: list[LidarrQueueItem] = field(default_factory=list[LidarrQueueItem])

    def __post_init__(self):
        """Post init."""
        self.records = [LidarrQueueItem(record) for record in self.records]


@dataclass(init=False)
class LidarrRelease(_ReleaseCommon):
    """Lidarr release attributes."""

    albumTitle: str
    artistName: str
    discography: bool
    preferredWordScore: int
    quality: type[_Quality] = field(default=_Quality)
    releaseHash: str

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.quality = _Quality(self.quality)


@dataclass(init=False)
class LidarrRename(_Rename):
    """Lidarr rename attributes."""

    albumId: int
    artistId: int
    trackFileId: int
    trackNumbers: list[int]


@dataclass(init=False)
class LidarrRetag(LidarrRename):
    """Lidarr retag attributes."""

    changes: list[_RetagChange] | None = None
    path: str

    def __post_init__(self):
        """Post init."""
        self.changes = [_RetagChange(change) for change in self.changes or []]


@dataclass(init=False)
class LidarrSearch(BaseModel):
    """Lidarr search attributes."""

    artist: type[LidarrArtist] = field(default=LidarrArtist)
    foreignId: str
    id: int

    def __post_init__(self):
        """Post init."""
        self.artist = LidarrArtist(self.artist)


@dataclass(init=False)
class LidarrTagDetails(_TagDetails):
    """Lidarr tag details attributes."""

    artistIds: list[int]


@dataclass(init=False)
class LidarrTrack(LidarrRename):
    """Lidarr track attributes."""

    absoluteTrackNumber: int
    artist: type[LidarrArtist] = field(default=LidarrArtist)
    duration: int
    explicit: bool
    hasFile: bool
    id: int
    mediumNumber: int
    ratings: type[_Ratings] = field(default=_Ratings)
    title: str
    trackNumber: int

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.artist = LidarrArtist(self.artist)
        self.ratings = _Ratings(self.ratings)


@dataclass(init=False)
class LidarrTrackFile(_LidarrMediaInfo_Quality, _LidarrCommon):
    """Lidarr track file attributes."""

    artistId: int
    audioTags: type[_LidarrAudioTags] = field(default=_LidarrAudioTags)
    dateAdded: datetime
    path: str
    qualityWeight: int
    size: int

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.audioTags = _LidarrAudioTags(self.audioTags)


@dataclass(init=False)
class LidarrTrackFileEditor(BaseModel):
    """Lidarr track file attributes."""

    quality: type[_Quality] = field(default=_Quality)
    trackFileIds: list[int]

    def __post_init__(self):
        """Post init."""
        self.quality = _Quality(self.quality)


@dataclass(init=False)
class _LidarrParsedAlbumInfo(BaseModel):
    """Lidarr parsed album info attributes."""

    albumTitle: str
    artistName: str
    artistTitleInfo: type[_LidarrArtistTitleInfo] = field(
        default=_LidarrArtistTitleInfo
    )
    quality: type[_Quality] = field(default=_Quality)
    releaseDate: int | datetime
    discography: bool
    discographyStart: int
    discographyEnd: int
    releaseGroup: str
    releaseHash: str
    releaseVersion: str

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.artistTitleInfo = _LidarrArtistTitleInfo(self.artistTitleInfo)
        self.quality = _Quality(self.quality)


@dataclass(init=False)
class LidarrParse(BaseModel):
    """Lidarr parse attributes."""

    albums: list[LidarrAlbum] | None = None
    artist: type[LidarrArtist] = field(default=LidarrArtist)
    parsedAlbumInfo: type[_LidarrParsedAlbumInfo] = field(
        default=_LidarrParsedAlbumInfo
    )
    title: str

    def __post_init__(self):
        """Post init."""
        self.albums = [LidarrAlbum(album) for album in self.albums or []]
        self.artist = LidarrArtist(self.artist)
        self.parsedAlbumInfo = _LidarrParsedAlbumInfo(self.parsedAlbumInfo)


@dataclass(init=False)
class LidarrArtistEditor(_Editor):
    """Lidarr artist editor attributes."""

    artistIds: list[int]


@dataclass(init=False)
class LidarrManualImport(_ManualImport):
    """Lidarr manual import attributes."""

    additionalFile: bool
    album: type[LidarrAlbum] = field(default=LidarrAlbum)
    albumReleaseId: int
    artist: type[_LidarrArtist] = field(default=LidarrArtist)
    audioTags: type[_LidarrAudioTags] = field(default=_LidarrAudioTags)
    disableReleaseSwitching: bool
    replaceExistingFiles: bool
    tracks: list[LidarrTrack] | None = None

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.album = LidarrAlbum(self.album)
        self.artist = LidarrArtist(self.artist)
        self.audioTags = _LidarrAudioTags(self.audioTags)
        self.tracks = [LidarrTrack(track) for track in self.tracks or []]


@dataclass(init=False)
class LidarrMetadataProvider(BaseModel):
    """Lidarr metadata provider attributes."""

    metadataSource: str
    writeAudioTags: str
    scrubAudioTags: bool
    id: int


@dataclass(init=False)
class LidarrAlbumStudio(BaseModel):
    """Lidarr album studio attributes."""

    artist: list[_Monitor] | None = None
    monitoringOptions: type[_MonitorOption] = field(default=_MonitorOption)

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.artist = [_Monitor(x) for x in self.artist or []]
        self.monitoringOptions = _MonitorOption(self.monitoringOptions)



"""Sonarr API."""
from __future__ import annotations

from datetime import datetime
from typing import Any

from aiohttp.client import ClientSession

from .const import (
    ALL,
    DATE,
    EPISODE_ID,
    EVENT_TYPE,
    IS_VALID,
    NOTIFICATION,
    PAGE,
    PAGE_SIZE,
    PATH,
    SERIES_ID,
    SORT_DIRECTION,
    SORT_KEY,
    TERM,
    TITLE,
    HTTPMethod,
)
from .exceptions import ArrException
from .models.host_configuration import PyArrHostConfiguration
from .models.request import Command, SortDirection
from .models.sonarr import (
    SonarrBlocklist,
    SonarrCalendar,
    SonarrCommands,
    SonarrEpisode,
    SonarrEpisodeFile,
    SonarrEpisodeHistory,
    SonarrEpisodeMonitor,
    SonarrEventType,
    SonarrHistory,
    SonarrImportList,
    SonarrLanguage,
    SonarrManualImport,
    SonarrNamingConfig,
    SonarrNotification,
    SonarrParse,
    SonarrQueue,
    SonarrQueueDetail,
    SonarrRelease,
    SonarrRename,
    SonarrSeasonPass,
    SonarrSeries,
    SonarrSeriesAdd,
    SonarrSeriesLookup,
    SonarrSortKeys,
    SonarrTagDetails,
    SonarrWantedMissing,
)
from .request_client import RequestClient


class SonarrClient(RequestClient):  # pylint: disable=too-many-public-methods
    """API client for Sonarr endpoints."""

    __name__ = "Sonarr"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host_configuration: PyArrHostConfiguration | None = None,
        session: ClientSession | None = None,
        hostname: str | None = None,
        ipaddress: str | None = None,
        url: str | None = None,
        api_token: str | None = None,
        port: int = 8989,
        ssl: bool | None = None,
        verify_ssl: bool | None = None,
        base_api_path: str | None = None,
        request_timeout: float = 30,
        raw_response: bool = False,
        api_ver: str = "v3",
    ) -> None:
        """Initialize Sonarr API."""
        super().__init__(
            port,
            request_timeout,
            raw_response,
            api_ver,
            host_configuration,
            session,
            hostname,
            ipaddress,
            url,
            api_token,
            ssl,
            verify_ssl,
            base_api_path,
        )

    async def async_get_episode_files(
        self, entryid: int, series: bool = False
    ) -> SonarrEpisodeFile | list[SonarrEpisodeFile]:
        """Get information about an episode file.

        episodeId: Get episode files matching id.
        series: Make True to search by series.
        """
        return await self._async_request(
            f"episodefile{'' if series else f'/{entryid}'}",
            params={SERIES_ID: entryid} if series else None,
            datatype=SonarrEpisodeFile,
        )

    async def async_get_queue(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 20,
        sort_dir: SortDirection = SortDirection.DEFAULT,
        sort_key: SonarrSortKeys = SonarrSortKeys.TIMELEFT,
        include_unknown_series_items: bool = False,
        include_series: bool = False,
        include_episode: bool = False,
    ) -> SonarrQueue:
        """Get information about download queue.

        page: Page number to return.
        page_size: Number of items per page.
        id: Filter to a specific episode ID.
        """
        params = {
            PAGE: page,
            PAGE_SIZE: page_size,
            SORT_DIRECTION: sort_dir.value,
            SORT_KEY: sort_key.value,
            "includeUnknownSeriesItems": str(include_unknown_series_items),
            "includeSeries": str(include_series),
            "includeEpisode": str(include_episode),
        }
        return await self._async_request("queue", params=params, datatype=SonarrQueue)

    async def async_get_queue_details(
        self,
        include_unknown_series_items: bool = False,
        include_series: bool = True,
        include_episode: bool = True,
    ) -> list[SonarrQueueDetail]:
        """Get details of all items in queue."""
        params = {
            "includeUnknownSeriesItems": str(include_unknown_series_items),
            "includeSeries": str(include_series),
            "includeEpisode": str(include_episode),
        }
        return await self._async_request(
            "queue/details",
            params=params,
            datatype=SonarrQueueDetail,
        )

    async def async_get_calendar(  # pylint: disable=too-many-arguments
        self,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        episodeid: int | None = None,
        include_series: bool = False,
        include_episode_file: bool = False,
        include_unmonitored: bool = False,
    ) -> SonarrCalendar | list[SonarrCalendar]:
        """Get upcoming episodes.

        If start/end not supplied episodes airing today and tomorrow will be returned
        episodeid: Specify to get calendar info from episode id
        """
        params = {}

        if start_date:
            params["start"] = start_date.strftime("%Y-%m-%d")
        if end_date:
            params["end"] = end_date.strftime("%Y-%m-%d")

        if include_series:
            params["includeSeries"] = "True"

        if include_episode_file:
            params["includeEpisodeFile"] = "True"

        if include_unmonitored:
            params["includeUnmonitored"] = "True"

        return await self._async_request(
            f"calendar{'' if episodeid is None else f'/{episodeid}'}",
            params=params,
            datatype=SonarrCalendar,
        )

    async def async_sonarr_command(  # pylint: disable=too-many-arguments
        self,
        command: SonarrCommands,
        clientid: int | None = None,
        copymode: bool = True,
        episodeids: list[int] | None = None,
        files: list[int] | None = None,
        path: str | None = None,
        season: int | None = None,
        seriesid: int | list[int] | None = None,
    ) -> Command:
        """Send a command to Sonarr.

        Specify clientid for DownloadedEpisodesScan (Optional)
        Specify episodeIds for EpisodeSearch
        Specify files for RenameFiles
        Specify path for DownloadedEpisodesScan (Optional)
        Specify season for SeasonSearch
        Specify seriesId for:
            RefreshSeries (Optional),
            RenameSeries (list[int]),
            RescanSeries (Optional),
            SeasonSearch,
            SeriesSearch
        """
        data: dict[str, str | int | list[int]] = {"name": command.value}
        if clientid is not None:
            data["downloadClientId"] = clientid
        if episodeids is not None:
            data["episodeIds"] = episodeids
        if files is not None:
            data["files"] = files
        if path is not None:
            data[PATH] = path
        if seriesid is not None:
            if command == SonarrCommands.RENAME_SERIES:
                data["seriesIds"] = seriesid
            else:
                data[SERIES_ID] = seriesid
        if command is SonarrCommands.DOWNLOADED_EPISODES_SCAN:
            data["importMode"] = "Copy" if copymode else "Move"
        if season is not None:
            data["seasonNumber"] = season
        return await self._async_request(
            "command",
            data=data,
            datatype=Command,
            method=HTTPMethod.POST,
        )

    async def async_get_episodes(
        self, entryid: int, series: bool = False
    ) -> SonarrEpisode | list[SonarrEpisode]:
        """Get all episodes from a given series or episode id."""
        return await self._async_request(
            f"episode{'' if series else f'/{entryid}'}",
            params={SERIES_ID: entryid} if series else None,
            datatype=SonarrEpisode,
        )

    async def async_edit_episode(self, data: SonarrEpisode) -> SonarrEpisode:
        """Edit given episodes, currently only monitored is changed."""
        return await self._async_request(
            f"episode/{data.id}",
            data=data,
            datatype=SonarrEpisode,
            method=HTTPMethod.PUT,
        )

    async def async_edit_episode_file_quality(
        self, data: SonarrEpisodeFile
    ) -> SonarrEpisodeFile:
        """Edit an episode file quality.

        Quality only works, currently.
        Quality.id and Revision.version
        """
        return await self._async_request(
            "episodefile",
            data=data,
            datatype=SonarrEpisodeFile,
            method=HTTPMethod.PUT,
        )

    async def async_delete_episode_file(self, fileid: int) -> None:
        """Delete the episode file with corresponding id."""
        return await self._async_request(
            f"episodefile/{fileid}", method=HTTPMethod.DELETE
        )

    async def async_get_history(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 10,
        sort_key: SonarrSortKeys = SonarrSortKeys.DATE,
        recordid: int | None = None,
        event_type: SonarrEventType | None = None,
    ) -> SonarrHistory:
        """Get history (grabs/failures/completed).

        Args:
            sort_key: series.title, id, seriesid, episodeid, date,
                    quality, or sourcetitle (others do not apply)
            page: Page number to return.
            page_size: Number of items per page.
            id: Filter to a specific episode ID.
        """
        params = {
            PAGE: page,
            PAGE_SIZE: page_size,
            SORT_KEY: sort_key.value,
        }
        if event_type and event_type in SonarrEventType:
            params[EVENT_TYPE] = event_type.value
        if recordid is not None:
            params[EPISODE_ID] = recordid
        return await self._async_request(
            "history", datatype=SonarrHistory, params=params
        )

    async def async_get_history_since(
        self,
        date: datetime | None = None,
        seriesid: int | None = None,
        event_type: SonarrEventType | None = None,
    ) -> list[SonarrEpisodeHistory]:
        """Get history since specified date.

        seriesid: include to search history by series id (date will not apply)
        """
        if date is None and seriesid is None:
            raise ArrException(self, "Either date or seriesid is required")
        params: dict[str, int | str] = {}
        if isinstance(date, datetime):
            params[DATE] = date.strftime("%Y-%m-%d")
        elif seriesid is not None:
            params[SERIES_ID] = seriesid
        if event_type and event_type in SonarrEventType:
            params[EVENT_TYPE] = event_type.value
        return await self._async_request(
            f"history/{'since' if isinstance(date, datetime) else 'series'}",
            params=params,
            datatype=SonarrEpisodeHistory,
        )

    # POST history/failed

    async def async_get_wanted(  # pylint: disable=too-many-arguments
        self,
        page: int = 1,
        page_size: int = 10,
        sort_key: SonarrSortKeys = SonarrSortKeys.AIR_DATE_UTC,
        sort_dir: SortDirection = SortDirection.DEFAULT,
        include_series: bool = False,
    ) -> SonarrWantedMissing:
        """Get missing episode (episodes without files).

        Args:
            sort_key: series.title, id, seriesid, airdateutc, ratings, or title
            (others do not apply)
            page: Page number to return.
            page_size: Number of items per page.
        """
        params = {
            SORT_KEY: sort_key.value,
            PAGE: page,
            PAGE_SIZE: page_size,
            SORT_DIRECTION: sort_dir.value,
        }

        if include_series:
            params["includeSeries"] = "True"

        return await self._async_request(
            "wanted/missing",
            params=params,
            datatype=SonarrWantedMissing,
        )

    async def async_parse_title_or_path(
        self, title: str | None = None, path: str | None = None
    ) -> SonarrParse:
        """Return the result of parsing a title or path.

        series and episodes will be
        returned only if the parsing matches to a specific series and one or more
        episodes. series and episodes will be formatted the same as Series and Episode
        responses.

        title: Title of series / episode
        path: file path of series / episode
        """
        if title is None and path is None:
            raise ArrException(message="A title or path must be specified")
        params = {}
        if title is not None:
            params[TITLE] = title
        if path is not None:
            params[PATH] = path
        return await self._async_request("parse", params=params, datatype=SonarrParse)

    async def async_get_release(
        self, episodeid: int | None = None
    ) -> list[SonarrRelease]:
        """Query indexers for latest releases.

        episodeid: included in API docs but does not seem to do anything
        """
        return await self._async_request(
            "release",
            params=None if episodeid is None else {EPISODE_ID: episodeid},
            datatype=SonarrRelease,
        )

    async def async_download_release(self, data: SonarrRelease) -> SonarrRelease:
        """Add a previously searched release to the download client.

        If the release is
        still in the search cache (30 minute cache). If the release is not found
        in the cache it will return a 404.

        Only the guid and indexerId are required
        but a full SonarrRelease object will work
        """
        return await self._async_request(
            "release",
            data=data,
            datatype=SonarrRelease,
            method=HTTPMethod.POST,
        )

    async def async_push_release(self, data: SonarrRelease) -> list[SonarrRelease]:
        """Add a release.

        If the title is wanted, Sonarr will grab it.
        """
        return await self._async_request(
            "release/push",
            data=data,
            datatype=SonarrRelease,
            method=HTTPMethod.POST,
        )

    async def async_get_series(
        self, seriesid: int | None = None
    ) -> SonarrSeries | list[SonarrSeries]:
        """Return all series in your collection or the series with the matching.

        series ID if one is found.
        """
        return await self._async_request(
            f"series{'' if seriesid is None else f'/{seriesid}'}",
            datatype=SonarrSeries,
        )

    async def async_add_series(self, data: SonarrSeriesAdd) -> SonarrSeries:
        """Add a new series to your collection."""
        return await self._async_request(
            "series",
            data=data,
            datatype=SonarrSeries,
            method=HTTPMethod.POST,
        )

    async def async_edit_series(self, data: SonarrSeries) -> SonarrSeries:
        """Edit an existing series."""
        return await self._async_request(
            "series",
            data=data,
            datatype=SonarrSeries,
            method=HTTPMethod.PUT,
        )

    async def async_delete_series(
        self, seriesid: int, delete_files: bool = False
    ) -> dict:
        """Delete the series with the given id.

        delete_files: If true series folder and files will be deleted.
        """
        return await self._async_request(
            f"series/{seriesid}",
            params={"deleteFiles": str(delete_files)},
            method=HTTPMethod.DELETE,
        )

    # POST series/import

    async def async_lookup_series(
        self, term: str | None = None, seriesid: int | None = None
    ) -> list[SonarrSeriesLookup]:
        """Search for new shows on TheTVDB.com utilizing sonarr.tv's.

        caching and augmentation proxy.
        """
        if term is None and seriesid is None:
            raise ArrException(self, "A term or TVDB id must be included")
        return await self._async_request(
            "series/lookup",
            datatype=SonarrSeriesLookup,
            params={TERM: term if term else f"tvdb:{seriesid}"},
        )

    async def async_get_import_lists(
        self, listid: int | None = None
    ) -> SonarrImportList | list[SonarrImportList]:
        """Get information about import list.

        id: Get import list matching id. Leave blank for all.
        """
        return await self._async_request(
            f"importlist{'' if listid is None else f'/{listid}'}",
            datatype=SonarrImportList,
        )

    async def async_edit_import_list(self, data: SonarrImportList) -> SonarrImportList:
        """Edit import list."""
        return await self._async_request(
            "importlist", data=data, datatype=SonarrImportList, method=HTTPMethod.PUT
        )

    async def async_add_import_list(self, data: SonarrImportList) -> SonarrImportList:
        """Add import list."""
        return await self._async_request(
            "importlist", data=data, datatype=SonarrImportList, method=HTTPMethod.POST
        )

    async def async_test_import_lists(
        self, data: SonarrImportList | None = None
    ) -> bool:
        """Test all import lists."""
        _res = await self._async_request(
            f"importlist/test{ALL if data is None else ''}",
            data=None if data is None else data,
            method=HTTPMethod.POST,
        )
        if data is None:
            for item in _res:
                if item[IS_VALID] is False:
                    return False
        return True

    async def async_season_pass(self, data: SonarrSeasonPass) -> None:
        """Change monitoring status via season pass."""
        return await self._async_request(
            "seasonPass", data=data, method=HTTPMethod.POST
        )

    async def async_episode_monitor(
        self, episodeids: list[int], monitored: bool = True
    ) -> list[SonarrEpisodeMonitor]:
        """Change monitoring status via season pass."""
        return await self._async_request(
            "episode/monitor",
            data={"episodeIds": episodeids, "monitored": str(monitored)},
            datatype=SonarrEpisodeMonitor,
            method=HTTPMethod.PUT,
        )

    async def async_get_languages(  # type: ignore[override]
        self, langid: int | None = None
    ) -> SonarrLanguage | list[SonarrLanguage]:
        """Get language profiles."""
        return await self._async_request(
            f"languageprofile{'' if langid is None else f'/{langid}'}",
            datatype=SonarrLanguage,
        )

    async def async_edit_language(self, data: SonarrLanguage) -> SonarrLanguage:
        """Edit language profiles."""
        return await self._async_request(
            "languageprofile",
            data=data,
            datatype=SonarrLanguage,
            method=HTTPMethod.PUT,
        )

    async def async_delete_language(self, languageid: int) -> None:
        """Delete language profiles."""
        return await self._async_request(
            f"languageprofile/{languageid}",
            datatype=SonarrLanguage,
            method=HTTPMethod.DELETE,
        )

    async def async_add_language(self, data: SonarrLanguage) -> SonarrLanguage:
        """Add language profiles."""
        return await self._async_request(
            "languageprofile",
            data=data,
            datatype=SonarrLanguage,
            method=HTTPMethod.POST,
        )

    async def async_get_blocklist(
        self,
        page: int = 1,
        page_size: int = 10,
        sort_dir: SortDirection = SortDirection.DEFAULT,
        sort_key: SonarrSortKeys = SonarrSortKeys.DATE,
    ) -> SonarrBlocklist:
        """Return blocklisted releases.

        Args:
            page: Page to be returned.
            page_size: Number of results per page.
            sort_key: series.title, id, seriesid, date, indexer,
                    message, quality, or sourcetitle (Others do not apply)
        """
        params = {
            PAGE: page,
            PAGE_SIZE: page_size,
            SORT_DIRECTION: sort_dir.value,
            SORT_KEY: sort_key.value,
        }
        return await self._async_request(
            "blocklist",
            params=params,
            datatype=SonarrBlocklist,
        )

    async def async_get_naming_config(self) -> SonarrNamingConfig:
        """Get information about naming configuration."""
        return await self._async_request("config/naming", datatype=SonarrNamingConfig)

    async def async_edit_naming_config(
        self, data: SonarrNamingConfig
    ) -> SonarrNamingConfig:
        """Edit Settings for file and folder naming."""
        return await self._async_request(
            "config/naming",
            data=data,
            datatype=SonarrNamingConfig,
            method=HTTPMethod.PUT,
        )

    async def async_get_notifications(
        self, notifyid: int | None = None
    ) -> SonarrNotification | list[SonarrNotification]:
        """Get information about notification.

        id: Get notification matching id. Leave blank for all.
        """
        return await self._async_request(
            f"notification{'' if notifyid is None else f'/{notifyid}'}",
            datatype=SonarrNotification,
        )

    async def async_edit_notification(
        self, data: SonarrNotification
    ) -> SonarrNotification:
        """Edit a notification."""
        return await self._async_request(
            NOTIFICATION,
            data=data,
            datatype=SonarrNotification,
            method=HTTPMethod.PUT,
        )

    async def async_add_notification(
        self, data: SonarrNotification
    ) -> SonarrNotification:
        """Add a notification."""
        return await self._async_request(
            NOTIFICATION,
            data=data,
            datatype=SonarrNotification,
            method=HTTPMethod.POST,
        )

    async def async_test_notifications(
        self, data: SonarrNotification | None = None
    ) -> bool:
        """Test a notification configuration."""
        _res = await self._async_request(
            f"notification/test{ALL if data is None else ''}",
            data=None if data is None else data,
            method=HTTPMethod.POST,
        )
        if data is None:
            for item in _res:
                if item[IS_VALID] is False:
                    return False
        return True

    async def async_get_rename(self, seriesid: int) -> list[SonarrRename]:
        """Get files matching specified id that are not properly renamed yet."""
        return await self._async_request(
            "rename",
            params={SERIES_ID: seriesid},
            datatype=SonarrRename,
        )

    async def async_get_manual_import(
        self,
        downloadid: str,
        folder: str | None = None,
        filterexistingfiles: bool = True,
    ) -> list[SonarrManualImport]:
        """Get manual import."""
        params = {
            "downloadId": downloadid,
            "filterExistingFiles": str(filterexistingfiles),
            "folder": folder if folder is not None else "",
        }
        return await self._async_request(
            "manualimport", params=params, datatype=SonarrManualImport
        )

    async def async_edit_manual_import(
        self, data: SonarrManualImport
    ) -> list[SonarrManualImport]:
        """Get manual import."""
        return await self._async_request(
            "manualimport",
            data=data,
            datatype=SonarrManualImport,
            method=HTTPMethod.PUT,
        )

    async def async_get_tags_details(
        self, tagid: int | None = None
    ) -> SonarrTagDetails | list[SonarrTagDetails]:
        """Get information about tag details.

        id: Get tag details matching id. Leave blank for all.
        """
        return await self._async_request(
            f"tag/detail{'' if tagid is None else f'/{tagid}'}",
            datatype=SonarrTagDetails,
        )

    async def async_get_localization(self) -> Any:
        """Get localization strings."""
        raise NotImplementedError()

    async def async_delete_metadata_profile(self, profileid: int) -> Any:
        """Delete a metadata profile."""
        raise NotImplementedError()



"""Sonarr Common Models. These are only for internal module use."""
# pylint: disable=invalid-name, too-many-instance-attributes, too-few-public-methods
from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime

from .base import BaseModel, get_datetime
from .request_common import (
    _Common3,
    _Common4,
    _Common5,
    _Common6,
    _Common9,
    _CommonAttrs,
    _HistoryCommon,
    _Quality,
    _QualityCommon,
    _Ratings,
    _TitleInfo,
)


@dataclass(init=False)
class _SonarrSeriesAlternateTitle(BaseModel):
    """Sonarr series alternate titles attributes."""

    seasonNumber: int
    title: str


@dataclass(init=False)
class _SonarrCommon2(_SonarrSeriesAlternateTitle, _Common6):
    """Sonarr common attributes."""

    airDate: datetime
    airDateUtc: datetime
    episodeFileId: int
    episodeNumber: int
    hasFile: bool
    id: int
    seriesId: int


@dataclass(init=False)
class _SonarrQualityProfileValueAttr(_Common3):
    """Sonarr quality profile value attributes."""

    weight: int


@dataclass(init=False)
class _SonarrImages(_Common5):
    """Sonarr images attributes."""

    remoteUrl: str


@dataclass(init=False)
class _SonarrSeries2(_Common6, _Common9):
    """Sonarr parse attributes."""

    added: datetime
    airTime: str
    cleanTitle: str
    ended: bool
    firstAired: datetime
    folder: str
    id: int
    images: list[_SonarrImages] | None = None
    languageProfileId: int
    lastInfoSync: datetime
    network: str
    path: str
    qualityProfileId: int
    ratings: type[_Ratings] = field(default=_Ratings)
    seasonFolder: bool
    seasons: list[_SonarrSeriesSeason] | None = None
    seriesType: str
    sortTitle: str
    status: str
    tags: list[int]
    titleSlug: int
    tvdbId: int
    tvMazeId: int
    tvRageId: int
    useSceneNumbering: bool

    def __post_init__(self):
        self.images = [_SonarrImages(image) for image in self.images or []]
        self.ratings = _Ratings(self.ratings)
        self.seasons = [_SonarrSeriesSeason(season) for season in self.seasons or []]


@dataclass(init=False)
class _SonarrEpisodeFile(_QualityCommon):
    """Sonarr episode file attributes."""

    dateAdded: datetime
    id: int
    language: type[_SonarrQualityProfileValueAttr] = field(
        default=_SonarrQualityProfileValueAttr
    )
    languageCutoffNotMet: bool
    mediaInfo: type[_CommonAttrs] = field(default=_CommonAttrs)
    path: str
    relativePath: str
    releaseGroup: str
    seasonNumber: int
    seriesId: int
    size: int

    def __post_init__(self):
        super().__post_init__()
        self.language = _SonarrQualityProfileValueAttr(self.language)
        self.mediaInfo = _CommonAttrs(self.mediaInfo)


@dataclass(init=False)
class _SonarrEpisodeMonitor(_SonarrCommon2):
    """Sonarr common attributes."""

    absoluteEpisodeNumber: int
    unverifiedSceneNumbering: bool


@dataclass(init=False)
class _SonarrCommon(_SonarrEpisodeMonitor):
    """Sonarr common attributes."""

    episodeFile: type[_SonarrEpisodeFile] = field(default=_SonarrEpisodeFile)

    def __post_init__(self):
        super().__post_init__()
        self.episodeFile = _SonarrEpisodeFile(self.episodeFile)


@dataclass(init=False)
class _SonarrEpisodeHistoryData(_Common4, _HistoryCommon):
    """Sonarr history record data attributes."""

    approved: bool
    downloadAllowed: bool
    downloadClientName: str
    droppedPath: str
    fileId: int
    guid: str
    importedPath: str
    indexerId: int
    preferredWordScore: int
    qualityWeight: int
    rejections: list[str]
    sceneSource: bool
    seasonNumber: int
    title: str
    tvdbId: str
    tvRageId: int


@dataclass(init=False)
class _SonarrSeasonStatistics(BaseModel):
    """Sonarr season statistics attributes."""

    episodeCount: int
    episodeFileCount: int
    percentOfEpisodes: float
    previousAiring: str | None = None
    seasonCount: int
    sizeOnDisk: int
    totalEpisodeCount: int

    def __post_init__(self):
        self.previousAiring = get_datetime(self.previousAiring)


@dataclass(init=False)
class _SonarrSeriesSeason(BaseModel):
    """Sonarr wanted missing series season attributes."""

    monitored: bool
    seasonNumber: int
    statistics: type[_SonarrSeasonStatistics] = field(default=_SonarrSeasonStatistics)

    def __post_init__(self):
        self.statistics = _SonarrSeasonStatistics(self.statistics)


@dataclass(init=False)
class _SonarrSeriesCommon(_SonarrSeries2):
    """Sonarr series common attributes."""

    images: list[_SonarrImages] | None = None
    remotePoster: str
    seasons: list[_SonarrSeriesSeason] | None = None
    statistics: type[_SonarrSeasonStatistics] = field(default=_SonarrSeasonStatistics)

    def __post_init__(self):
        self.images = [_SonarrImages(image) for image in self.images or []]
        self.ratings = _Ratings(self.ratings)
        self.seasons = [_SonarrSeriesSeason(season) for season in self.seasons or []]
        self.statistics = _SonarrSeasonStatistics(self.statistics)


@dataclass(init=False)
class _SonarrWantedMissingRecord(_SonarrCommon):
    """Sonarr wanted missing record attributes."""

    downloading: bool
    sceneEpisodeNumber: int
    sceneSeasonNumber: int
    series: type[_SonarrSeries2] = field(default=_SonarrSeries2)
    tvDbEpisodeId: int

    def __post_init__(self):
        """Post init."""
        super().__post_init__()
        self.series = _SonarrSeries2(self.series)


@dataclass(init=False)
class _SonarrParseEpisodeInfo(BaseModel):
    """Sonarr parse episode info attributes."""

    absoluteEpisodeNumbers: list[int]
    episodeNumbers: list[int]
    fullSeason: bool
    isAbsoluteNumbering: bool
    isDaily: bool
    isMultiSeason: bool
    isPartialSeason: bool
    isPossibleSceneSeasonSpecial: bool
    isPossibleSpecialEpisode: bool
    isSeasonExtra: bool
    language: type[_Common3] = field(default=_Common3)
    quality: type[_Quality] = field(default=_Quality)
    releaseGroup: str
    releaseHash: str
    releaseTitle: str
    releaseTokens: str
    seasonNumber: int
    seasonPart: int
    seriesTitle: str
    seriesTitleInfo: type[_TitleInfo] = field(default=_TitleInfo)
    special: bool
    specialAbsoluteEpisodeNumbers: list[int]

    def __post_init__(self):
        self.language = _Common3(self.language)
        self.quality = _Quality(self.quality)
        self.seriesTitleInfo = _TitleInfo(self.seriesTitleInfo)


@dataclass(init=False)
class _SonarrAddOptions(BaseModel):
    """Sonarr add options attributes."""

    ignoreEpisodesWithFiles: bool | None = None
    ignoreEpisodesWithoutFiles: bool | None = None
    searchForMissingEpisodes: bool | None = None


@dataclass(init=False)
class _SonarrLanguageItem(BaseModel):
    """Sonarr language item attributes."""

    language: type[_Common3] = field(default=_Common3)
    allowed: bool

    def __post_init__(self):
        self.language = _Common3(self.language)
